#!/usr/bin/perl

# cttel.pl
#   Mac/Linux Connect Tunnel telnet frontend.

use strict;
use warnings;

use Socket;
use Pod::Usage;
use Getopt::Long;
use File::Copy;

use constant XG_SOCK_FILE => '/var/run/Aventail/AvConnect.sock';

# May be modified by packager to point at the real directory.
use constant XG_INST_DIR  => '/usr/local/Aventail';
use constant XG_SWING     => XG_INST_DIR . '/nui/nui.jar';

use constant XG_AVC       => XG_INST_DIR . '/AvConnect';
use constant XG_MAN       => XG_INST_DIR . '/man/ct.5';
use constant XG_REMOTE_READ => 400;

use constant XG_USER_HOME   => (getpwuid($<))[7];
use constant XG_MACOS       => ($^O eq 'darwin');
use constant XG_CONFIG_DIR  => 
    XG_MACOS?
        (XG_USER_HOME . '/Library/Application Support/Aventail')
      : (XG_USER_HOME . '/.aventail');
use constant XG_GLOBAL_CONFIG_DIR =>
    XG_MACOS?
        '/Library/Application Support/Aventail'
      : XG_INST_DIR;

use constant XG_CONFIG_DIR_PARAM    => '--avconf-dir=' . XG_CONFIG_DIR;
use constant XG_PROXY_CONFIG_FILE   => '.AvConnect.pcfg';
use constant XG_PROXY_CONFIG        =>
                                XG_CONFIG_DIR . '/' . XG_PROXY_CONFIG_FILE;
use constant XG_GLOBAL_PROXY_CONFIG =>
                                XG_GLOBAL_CONFIG_DIR . '/' .
                                                XG_PROXY_CONFIG_FILE;
use constant XG_GLOBAL_UI_CONFIG    => XG_GLOBAL_CONFIG_DIR . '/' .
                                                '.ctrc';
use constant XG_AVC_CONFIG          => XG_CONFIG_DIR . '/.avc.conf';
use constant XG_AVC_CONF_HOST       => 'REMOTE_HOST';

use constant XG_PROXY_PARAM         => '--use-proxy';
use constant XG_PROXY_PROG          => XG_INST_DIR . '/avpconf';


# Any constant defined as XGI_FOO => 'xyz.h' will be replaced with
# the value of the first constant named FOO in xyz.h.
use constant XGI_KEY_PROXY_PACE     => 'ProxyAutoConfigEnable';
use constant XGI_KEY_AUTO_PROXY     => 'ProxyAutoConfigURLString';
use constant XGI_KEY_PROXY_MODE     => 'ProxyConfigurationSelected';

use constant XGI_MSG_LN             => '--------------------------------------------';
use constant XGI_MSG_WLC            => '               AVENTAIL CONNECT             ';
use constant XGI_MSG_NO_HOSTNAME    => 'Error: Appliance name or IP address not specified.';
use constant XGI_PRM_XG_SRV         => 'Enter Appliance HOST/IP(Default[%s]):';

# Force autoflush on STDOUT. Not that we're doing that many buffered writes to
# STDOUT, but there were a few last time I checked.
local $| = 1;

my $PRODUCT = 'Aventail Connect';
my $LOGFILE = '/var/log/AvConnect.log';

my $XG_PROMPT = 'CONNECT CMD>';
my $XG_CONNECTED = 'VPN connection established.';

my @HELP_CMDS = qw/? help/;

# Older perls seem to use bash even for calls to system LIST. Kill BASH_ENV
# to bypass malicious user scripts on such perls.
delete $ENV{BASH_ENV};
delete $ENV{IFS};
delete $ENV{CDPATH};

# Forgive us today our sins... :-)
$ENV{PAGER} = 'less';

# Copy args list before Getopt can strip them from @ARGV.
my @args = map /^(.*)$/, @ARGV;

# If STDIN is not a tty, we cannot do an interactive login, so force an
# autologon.
unless (-t STDIN) {
    unshift @args, '--autologon';
}

my %options;
Getopt::Long::Configure( qw/bundling pass_through/ );
# GetOptions( \%options, "help|h", "man", "version" );

GetOptions( \%options, 
    "help|h", 
    "man", 
    "version",  
    "use-proxy", 
    "proxy-config",
    "background|daemon|b",
    "debug|d",
    "certificate|cert|c=s",
    "username|user|u=s",
    "password|pass|p=s",
    "host|appliance|a=s",
    "rc=s",
    "team-id|t=s",
    "realm|login-group|r|l:s",
    "verbose",
    "no-detach",
    "no-banner",
    "autologon",
    "use-proxy",
    "pac-proxy=s",
    "proxy-user=s",
    "proxy-pass=s",
    "proxy-auth=s",
    "avconf-dir=s"
    );

# --proxy-config implies --use-proxy
$options{'use-proxy'} = 1 if $options{'proxy-config'};

if (@ARGV && !$options{host} && $ARGV[0] !~ /^-/) {
    $options{host} = $ARGV[0];
}

if ($options{help}) {
    show_pod();
    exit 0;
}

if ($options{man}) {
    show_help();
    exit 0;
}

if ($options{version}) {
    show_version();
    exit 0;
}

if(!(-d XG_CONFIG_DIR)){
    mkdir XG_CONFIG_DIR, 0755;
}
pull_config();

if ($options{"use-proxy"}) {

    my $avpSt = 0;
    if ($^O eq "darwin"){
       $avpSt = osx_create_proxy_cfg();
    }
    elsif(!(-e XG_PROXY_CONFIG)
    	|| -z XG_PROXY_CONFIG 
    		|| $options{"proxy-config"}) {
       $avpSt = system(XG_PROXY_PROG);
    }
    if($avpSt != 0) {
      my $rc = $avpSt >> 8;
      exit ($rc? $rc : 127);
    }

    unshift @args, "--use-proxy";
    
    # If we have a pac file, figure out what proxy we need, prompting
    # the user for the appliance name if necessary, and inject the
    # proxy as a --pac-proxy arg (prefixed into @args). May also add an
    # --appliance= arg if we prompted for the appliance.
    
    find_proxy();
}


local $SIG{INT} = 
    sub {
        # Re-enable terminal echo in case we were in the middle
        # of a password prompt
        system "stty echo" if -t STDIN;
        die "\nTerminated.\n";
    };
local $SIG{TERM} = $SIG{INT};

startng();

my $sock = xg_listen_sock();
unless (defined $sock) {
    die "$PRODUCT failed to start. Please check $LOGFILE\n";
}
my $rc = telnet_loop('127.0.0.1', $sock);
exit $rc;

sub pull_config {
    if (!-f XG_AVC_CONFIG && -r XG_GLOBAL_UI_CONFIG) {
        my $rconf = read_config(XG_GLOBAL_UI_CONFIG);
        if ($$rconf{'vpn.server'}) {
            open my $outf, '>', XG_AVC_CONFIG;
            if ($outf) {
                print $outf XG_AVC_CONF_HOST . "=$$rconf{'vpn.server'}\n";
                close $outf;
            }
        }
    }
    
    if (!-f XG_PROXY_CONFIG && -r XG_GLOBAL_PROXY_CONFIG) {
        # Copy in the global config to our local config
        copy(XG_GLOBAL_PROXY_CONFIG, XG_PROXY_CONFIG);
    } 
}

sub need_proxy {
    # If a kind user has specified --pac-proxy, we don't need to
    # do a thing. This is unlikely to happen.
    return if $options{"pac-proxy"};

    # If we can't read the proxy config file, bail out now.
    return unless -r XG_PROXY_CONFIG;

    my $rconf = read_config( XG_PROXY_CONFIG );
    return $$rconf{+XGI_KEY_PROXY_MODE} &&
           ($$rconf{+XGI_KEY_PROXY_MODE} == 2 
           	|| ($$rconf{+XGI_KEY_PROXY_MODE} == 4)) &&
           $$rconf{+XGI_KEY_AUTO_PROXY};
}

sub get_appliance_name {
    # If we already have a hostname, we'll just use that.
    return untaint_hostname( $options{host} ) if $options{host};

    # If the user wants an autologon, and they haven't specified
    # a proxy, we make appropriate dying noises.
    if ($options{autologon}) {
        die XGI_MSG_NO_HOSTNAME . "\n" .
          "Please provide an appliance name as --appliance=<yourappliance>\n";
    }

    # Joy, this isn't an autologon, so we get to prompt.
    return prompt_appliance_name();
}

sub get_default_appliance_name {
    return unless -r XG_AVC_CONFIG;
    my $props = read_config( XG_AVC_CONFIG );
    return $$props{+XG_AVC_CONF_HOST};
}

sub untaint_hostname {
    my $host = shift;
    return '' unless $host;

    # Accept valid DNS names, and some invalid ones as well.
    my ($untainted) = $host =~ /^([a-zA-Z0-9._%\@\-:]+)/;
    return $untainted;
}

sub prompt_appliance_name {
    my $default = get_default_appliance_name();
    $default = '' unless defined $default;

    unless ($options{'no-banner'}) {
        print XGI_MSG_LN, "\n";
        print XGI_MSG_WLC, "\n";
        print XGI_MSG_LN, "\n";

        # We just showed the banner, so make sure we
        # don't show it again.
        unshift @args, '--no-banner';
    }

    printf XGI_PRM_XG_SRV, $default;
    chomp( my $appname = <STDIN> );
    $appname = $default unless $appname;

    # Untaint appname so we can pass it to AvConnect later
    $appname = untaint_hostname($appname);

    die XGI_MSG_NO_HOSTNAME . "\n" unless $appname;

    unshift @args, "--appliance=$appname";
    return $appname;
}

sub assert_have_java {
    my $err = shift;

    my $where = qx{which java 2>/dev/null};
    chomp $where if $where;

    die "$err\n" unless $where && -x $where;
}

sub osx_create_proxy_cfg {
    # On OSX use ProxyDetector instead of avpconf
    my $appliance = "porxy-detect";
    assert_have_java( "Java is required for PAC proxy support, "
               . "but is not available in your PATH ($ENV{PATH})." );

    my $pdetect = "com.aventail.xg.proxy.ProxyDetector";
    my @res = qx/java -cp "@{ [ XG_SWING ] }" $pdetect $appliance/;
    if ($? || !@res) {
        # Java failed. What to do?
        die "Unable to detect proxy: $pdetect error code " . ($? >> 8) . "\n";
    }
    return $?;
}

sub query_proxy {
    # We need to get an appliance name, prompting the user if
    # necessary. We will *not* prompt for the appliance name
    # if autologon is set.
    my $appliance = get_appliance_name();
    die XGI_MSG_NO_HOSTNAME . "\n" unless $appliance;

    assert_have_java( "Java is required for PAC proxy support, "
               . "but is not available in your PATH ($ENV{PATH})." );

    my $pdetect = "com.aventail.xg.proxy.ProxyDetector";
    my @res = qx/java -cp "@{ [ XG_SWING ] }" $pdetect $appliance/;
    if ($? || !@res) {
        # Java failed. What to do?
        die "Unable to detect proxy: $pdetect error code " . ($? >> 8) . "\n";
    }

    chomp( my $proxy = $res[0] );
    # How do we untaint meaningfully?
    ($proxy) = $proxy =~ /(.*)/;
    my @params = grep /\S/, (split ' ', $proxy);
    return @params;
}

sub read_config {
    my $file = shift;
    open my $inf, '<', $file or return +{ };
    my $text = do { local $/; <$inf> };
    close $inf;

    my $ws = "[\t ]";
    # Throw away comments
    $text =~ s/^$ws*#.*//m;

    # Pull out key value pairs and return them
    return +{ $text =~ /^$ws*(\S+)$ws*=$ws*(.*)$/mg  }
}

sub find_proxy {
    if (need_proxy()) {
        my @ppars = query_proxy();
        unshift @args, @ppars if @ppars;
    }
}

sub show_help {
    system "man " . XG_MAN . " 2>/dev/null";
}

sub show_version {
    my $ver = get_version();
    print "$PRODUCT ", ($ver? $ver : "?"), "\n";
}

sub get_version {
    open my $inf, '<', XG_INST_DIR . "/version" or return;
    my $ver = <$inf>;
    chomp $ver if defined $ver;
    return clean_version($ver);
}

sub clean_version {
    my $ver = shift;
    return $ver unless defined $ver;

    $ver =~ s/^(\d+\.)(\d)(\d)\d*\./$1$2.$3./;
    return $ver;
}

sub telnet_loop {
    my ($host, $port) = @_;
    my $iaddr = inet_aton($host) or die "Can't resolve $host: $!\n";
    my $saddr = sockaddr_in($port, $iaddr);   
    my $proto = getprotobyname('TCP');
    socket(my $tsock, PF_INET, SOCK_STREAM, $proto)
            or die "Couldn't open socket to $host:$port: $!\n";
    connect($tsock, $saddr) or die "Couldn't connect to $host:$port: $!\n";
    
    binmode STDIN;
    binmode $tsock;
    my $rc = telnet_read_loop($tsock);
    
    close($tsock);

    return $rc;
}

sub telnet_read_loop {
    my $tsock = shift;
    my $c;

    my $rin = '';
    vec($rin, fileno(STDIN), 1) = 1;
    vec($rin, fileno($tsock), 1) = 1;

    my ($rout, $eout);

    my $rollingbuf = '';
    my $input = '';

    my $masked = 0;
    my $xg_connected;
    my $seen_input;
    my $suppress_output;

    my $inputty = -t STDIN;

    # Ditch SIGPIPEs. We expect to see a few of those, and we don't really
    # care.
    local $SIG{PIPE} = sub { };

    while (1) {
        my $nfound = select($rout = $rin, undef, $eout = $rin, undef);
        die "Select failed: $!\n" unless $nfound;
        
        if (vec($rout, fileno(STDIN), 1)) {
            my $inkey; 
            my $res = sysread(STDIN, $inkey, 5);

            if ($res) {
                $input .= $inkey;
                if ($input =~ /\n/ && $inputty) {
                    $seen_input = 1;
                    send_input($tsock, \$input, $xg_connected);
                }
            }
            else {
                # STDIN appears closed; let's make sure we don't try
                # reading from it again.
                vec($rin, fileno(STDIN), 1) = 0;
            }
        }

        if (vec($rout, fileno($tsock), 1)) {
            my $remotech;
            my $res = sysread($tsock, $remotech, XG_REMOTE_READ);
            return ($xg_connected? 0 : 1) unless $res;

            $rollingbuf = substr($rollingbuf, XG_REMOTE_READ * 3) 
                        if length($rollingbuf) >= 10 * XG_REMOTE_READ;
            $rollingbuf .= $remotech;

            if ($rollingbuf =~ /\Q$XG_PROMPT/i ||
                $rollingbuf =~ /\Q$XG_CONNECTED/i)
            {
                $xg_connected = 1;
            }

            my $maskreq = "\xFF\xFB\x01";
            my $unmaskreq = "\xFF\xFC\x01\x0D";

            if ($masked && 
                length($rollingbuf) >= length($unmaskreq) &&
                substr($rollingbuf, 
                       length($rollingbuf) - length($unmaskreq), 
                       length($unmaskreq)) eq $unmaskreq) 
            {
                $masked = 0;
                # Enable terminal echo
                system "stty echo" if $inputty;
                # Acknowledge password unmask request
                syswrite $tsock, ("\x01" x 3);
            }

            if (!$masked && 
                length($rollingbuf) >= length($maskreq) &&
                substr($rollingbuf, 
                       length($rollingbuf) - length($maskreq), 
                       length($maskreq)) eq $maskreq) 
            {
                $masked = 1;
                # Suppress terminal echo
                system "stty -echo" if $inputty;
                # Acknowledge password mask request
                syswrite $tsock, ("\x01" x 3);

                if (!$inputty && $input) {
                    send_one_input($tsock, \$input);
                }
            }

            $remotech =~ tr/\x01\xFF\xFB\xFC//d;
            syswrite STDOUT, $remotech unless $suppress_output;
        }
    }
}

sub write_prompt_prelude {
    my ($text, $prompt) = @_;
    my @frags = split /\n/, $text;
    for my $frag (@frags) {
        return if $frag =~ /\Q$prompt/;
        syswrite STDOUT, "$frag\n";
    }
}

sub send_one_input {
    my ($tsock, $rin) = @_;
    my ($line) = $$rin =~ /^(.*?\n)/;
    if ($line) {
        $$rin =~ s/^(.*?\n)//;
        syswrite $tsock, $line;
    }
}

sub send_input {
    my ($tsock, $rin, $help) = @_;

    my @frags = split /\n/, $$rin, -1;
    $$rin = pop @frags;

    for my $frag (@frags) {
        if ($help && grep /^\Q$frag\E$/i, @HELP_CMDS) {
            show_help();
            # XXX: This is seriously lame
            syswrite STDOUT, $XG_PROMPT;
        }
        else {
            syswrite $tsock, "$frag\n";
        }
    }
}

sub xg_listen_sock {
    unless (-f XG_SOCK_FILE) {
        die "Unable to start $PRODUCT\n";
    }

    my $tries = 0;
    while (++$tries < 5) {
        open my $sockf, '<', XG_SOCK_FILE or next;
        my $sc = <$sockf>;
        close $sockf;

        next unless defined $sc;
        chomp $sc;

        # Untaint port number to connect to.
        ($sc) = $sc =~ /(\d+)/;
        return $sc if $sc > 0;
    }
    continue {
        # Sleep for 250ms
        select undef, undef, undef, 0.25;
    }
    return undef;
}

sub startng {
    my $modtime;
    $modtime = -M XG_SOCK_FILE if -f XG_SOCK_FILE;
    
    my @ngargs = ( XG_AVC, '-b', XG_CONFIG_DIR_PARAM, @args );
    system @ngargs;
    my $tries = 0;
    while ((!-f XG_SOCK_FILE || 
            ($modtime && $modtime == -M XG_SOCK_FILE))  
            && ++$tries <= 12)
    {
        # Sleep in 250ms intervals
        select undef, undef, undef, 0.25;
    }
}

sub show_pod {
    pod2usage(0);
}

__END__

=head1 NAME

startct - Starts Aventail Connect

=head1 SYNOPSIS

startct [options] [appliance]

 Options:
  --help     		Show usage summary.
  --man      		Show man page.
  --version  		Show the version of the installed Aventail Connect.
  -y         		Accept appliance certificate even if verification fails.
  -u <name>  		Username.
  -p <pw>    		Password. To prompt just skip this option,
  	     		or pipe in the password to stdin	
  -l <group> 		Login group.
  -d         		Log extra debugging information (in /var/log/AvConnect.log)
  --autologon		Automatically connect Aventail Connect and run in background.
  			To stop Aventail connect use command 'stopct'. 
  --use-proxy		Use proxy to Connect, On Linux if proxy is not configured, 
  			this option triggers proxy configuration.
  --proxy-user		Username for proxy login.
  --proxy-pass		Password for proxy login.
  --proxy-config	Configure proxy for Aventail Connect(Linux Only), implies '--use-proxy'
  --ncr=<option>	Network conflict resolution options in Split Tunnel mode, options are 
  			A - Use Administrator preference (Default)
                        L - Prefer local network resources
                        R - Prefer remote network resources
  
 E.g.  			
  Start Aventail Connect
  $ startct -y -r<realm> -u<username> -p<password> <app-ip/host>

  Start Aventail Connect in debug mode
  $ startct -y -r<realm> -u<username> -p<password> -d <app-ip/host>

  Start Aventail Connect to run in background
  $ startct -y -r<realm> -u<username> -p<password> --autologon <app-ip/host>

  Start Aventail Connect to run in background with proxy
  $ startct -y -r<"my Realm"> -u<username> -p<password> --autologon 
    --use-proxy --proxy-user=<proxyuser> --proxy-pass=<proxypassword> <app-ip/host>

  Stop Aventail Connect when running in background
  $ stopct
	
=cut
